const fetch = require('node-fetch');
const uuid = require('uuid/v4');

class User {
	constructor(model, providerName) {
		this.model = model;
	}
	register(options={}, callback) {
		/*
		** @param {object} options.uniqueKey required
		** @param {object} options.credentialKey required
		** @param {object} options.data required
		** @param {function} options.onSuccess required
		** @param {function} options.onError required
		*/
		return async (req, res, next)=>{
			try {
				let _where = {};
				_where[options.uniqueKey] = options.data[options.uniqueKey];
				let user = await this.model.findOne({where: _where});
				if (user) throw Error('Username taken already!');
				let userData = {};
				let userDataKeys = Object.keys(options.data);
				for (let i = 0; i < userDataKeys.length; i++) {
					userData[userDataKeys[i]] = options.data[userDataKeys[i]];
				}
				user = this.model.build(userData);
				user.salt = this.model.createSalt();
				user.hash = this.model.createHash(user.salt, options.data[options.credentialKey]);
				await user.save();
				options.onSuccess(user);
			} catch (err) {
				options.onError(err);
			}
		};
	}
	login(options={}, callback) {
		/*
		** @param {object} options.uniqueKey required
		** @param {object} options.credentialKey required
		** @param {object} options.data required
		** @param {function} options.onSuccess required
		** @param {function} options.onError required
		*/
		return async (req, res, next)=>{
			try {
				let _where = {};
				_where[options.uniqueKey] = options.data[options.uniqueKey];
				let user = await this.model.findOne({where: _where});
				if (!user) throw Error(`Invalid email and password!`);
				let hash = this.model.createHash(user.salt, options.data[options.credentialKey]);
				if (hash !== user.hash) throw Error(`Invalid email and password.`);
				options.onSuccess(user);
			} catch (err) {
				options.onError(err);
			}
		};
	}
	verify(callback) {
		return async (req, res, next)=>{
			try {
				let user = await this.model.findOne({
					where: {verificationCode: req.body.user.verificationCode},
				});
				if (!user) throw Error(`Invalid verification code!`);
				user.isVerified = true;
				await user.save();
				let jwt = user.generateJWT({
					attributes: ['id', 'email'],
				});
				callback(jwt);
			} catch (err) {
				return next(err);
			}
		}
	}
	createResetToken(options, callback) {
		/*
		** @param options.criteria string default "username"
		** @param options.target string required | where the reset token will be populated.
		** @param options.duration number required | the duration time of the token in millisecond.
		*/
		if (typeof options === 'function') callback = options;
		return async (req, res, next)=>{
			try {
				let criteria = options.criteria || 'username';
				let _where = {};
				_where[criteria] = req.body.user[criteria];
				let user = await this.model.findOne({where: _where});
				if (!user) throw Error(`Invalid ${options.criteria}.`);
				user[options.target] = uuid();
				user[`${options.target}Expires`] = Date.now() + options.duration;
				await user.save();
				callback(user);
			} catch (err) {
				return next(err);
			}
		};
	}
	resetPassword(options, callback) {
		if (typeof options === 'function') callback = options;
		return async (req, res, next)=>{
			try {
				let resetPasswordToken = req.body.user.resetPasswordToken;
				let user = await this.model.findOne({where: {resetPasswordToken}}); 
				if (!user) throw Error('Reset token is invalid or expired.');
				if (parseInt(user.resetPasswordTokenExpires) < Date.now()) throw Error('Reset token is invalid or expired.');
				user.salt = this.model.createSalt();
				user.hash = this.model.createHash(user.salt, req.body.user.password);
				user.resetPasswordToken = null;
				user.resetPasswordTokenExpires = null;
				await user.save();
				callback();
			} catch(err) {
				return next(err);
			}
		}
	}
	facebook(options, callback) {
		if (typeof options === 'function') callback = options;
		return async (req, res, next)=>{
			try {
				let accessToken = req.body.accessToken;
				console.log(`Access token: ${accessToken}`);
				let response = await fetch(`https://graph.facebook.com/me?fields=email,first_name,last_name&access_token=${accessToken}`);
				let data = await response.json();
				if (data.error) throw Error(data.error.message);
				/*
				** (1) check if the user is already registered in local DB.
				*/
				let user = await this.model.findOne({where: {providerId: data.id}});
				if (!user) {
					/*
					** (1-i) If no, create a new user data in the DB.
					*/
					console.log(`# Facebook user is not registered. Create a new account.`);
					let userData = {
						providerId: data.id,
						isVerified: true,
						provider: "facebook"
					};
					userData[options.criteria] = data.email;
					for (let i = 0; i < options.attributes.length; i++) {
						let attribute = options.attributes[i];
						userData[attribute] = data[attribute];
					}
					if (options.preDetermined) {
						for (let i = 0; i < options.preDetermined.length; i++) {
							let dataSet = options.preDetermined[i];
							userData[dataSet.name] = dataSet.value;
						}
					}
					user = this.model.build(userData);
					await user.save();
				}
				/*
				** (2) Feedback with jwt token through callback.
				*/
				console.log(`# Facebook user is found in the DB.`);
				callback(user);
			} catch (err) {
				return next(err);
			}
		};
	}
}

function sendResetTokenEmail(providerName, user, target) {
	let postman = new Postman(providerName);
	postman.sendEmail({
		to: user.email,
		subject: `Reset your password.`,
		html: `
<p>Hi ${user.email}, </p>
<p>You requested ${target}.</p>
<p>
Reset password token:
<br/>
<strong>${user[target]}</strong>
</p>
`,
	});
}

function sendVerificationEmail(providerName, user) {
	let postman = new Postman(providerName);
	postman.sendEmail({
		to: user.email,
		subject: `Verify your email.`,
		html: `
<p>Hi ${user.email}, </p>
<p>Welcome to ${providerName}! <p>
<p>You are required to input the given verification code to complete your application.</p>
<p>
Verification code:
<br/>
<strong>${user.verificationCode}</strong>
</p>
`,
	});
}

module.exports = User;
